""" You can kind of see this as the scope of `just` when you 'import just'
The following functions become available:
just.__project__
just.__version__
just.run
just.print_version
"""

import os

import just.txt as txt
import just.json_ as json
import just.newl as newl
import just.yaml_ as yaml
import just.csv_ as csv
import just.pickle_ as pickle
from just.path_ import make_path
from just.path_ import glob
from just.requests_ import get
from just.requests_ import post
from just.dir import mkdir
from just.log import log


__project__ = "just"
__version__ = "0.6.72"

EXT_TO_MODULE = {
    "html": txt,
    "py": txt,
    "txt": txt,
    "newl": newl,
    "json": json,
    "jsonl": json,
    "yaml": yaml,
    "yml": yaml,
    "csv": csv,
    "tsv": csv,
    "pickle": pickle,
    "pkl": pickle,
}


def reader(fname, no_exist, read_func_name, fallback_type, ignore_exceptions):
    fname = make_path(fname)
    if not os.path.isfile(fname) and no_exist is not None:
        return no_exist
    ext = fname.split(".")[-1] if "." in fname[-6:] else "txt"
    if ext not in EXT_TO_MODULE and fallback_type == "RAISE":
        raise TypeError("just does not yet cover '{}'".format(ext))
    reader_module = EXT_TO_MODULE.get(ext, None) or EXT_TO_MODULE[fallback_type]
    read_fn = getattr(reader_module, read_func_name)
    if ignore_exceptions is not None:
        try:
            return read_fn(fname)
        except ignore_exceptions:
            return None
    else:
        return read_fn(fname)


def read(fname, no_exist=None, fallback_type="RAISE", ignore_exceptions=None):
    return reader(fname, no_exist, "read", fallback_type, ignore_exceptions)


def multi_read(star_path, no_exist=None, fallback_type="RAISE", ignore_exceptions=None):
    return {x: read(x, no_exist, fallback_type, ignore_exceptions) for x in glob(star_path)}


def writer(obj, fname, mkdir_no_exist, skip_if_exist, write_func_name):
    fname = make_path(fname)
    if skip_if_exist and os.path.isfile(fname):  # pragma: no cover
        return False
    if mkdir_no_exist:
        dname = os.path.dirname(fname)
        if dname not in set([".", "..", ""]):
            mkdir(dname)
    ext = fname.split(".")[-1] if "." in fname[-6:] else "txt"
    writer_module = EXT_TO_MODULE[ext]
    write_fn = getattr(writer_module, write_func_name)
    return write_fn(obj, fname)


def write(obj, fname, mkdir_no_exist=True, skip_if_exist=False):
    return writer(obj, fname, mkdir_no_exist, skip_if_exist, "write")

# only supported for JSON Lines so far.


def append(obj, fname, mkdir_no_exist=True, skip_if_exist=False):
    return writer(obj, fname, mkdir_no_exist, skip_if_exist, "append")


def multi_write(obj, fname, mkdir_no_exist=True, skip_if_exist=False):
    if not isinstance(fname, list) or not isinstance(obj, list):  # pragma: no cover
        raise NotImplementedError("Only list of fnames + list of objects supported.")
    return [write(o, fn, mkdir_no_exist, skip_if_exist)
            for o, fn in zip(obj, fname)]


def iread(fname, no_exist=None, fallback_type="RAISE", ignore_exceptions=None):
    return reader(fname, no_exist, "iread", fallback_type, ignore_exceptions)


def iwrite(obj, fname, mkdir_no_exist=True, skip_if_exist=False):
    return writer(obj, fname, mkdir_no_exist, skip_if_exist, "iwrite")


def remove(fname, no_exist=None):
    fname = make_path(fname)
    if not os.path.isfile(fname) and no_exist is not None:
        return False
    os.remove(fname)
    return True


def rename(fname, extension, no_exist=None):
    fname = make_path(fname)
    if not os.path.isfile(fname) and no_exist is not None:
        return False
    os.rename(fname, fname + "." + extension)
    return True



try:
    import ujson as json
except ImportError:
    import json


def read(fn):
    if fn.endswith(".jsonl"):
        raise TypeError("JSON Newline format can only be read by iread")
    with open(fn) as f:
        return json.load(f)


def append(obj, fn):
    with open(fn, "a+") as f:
        f.write(json.dumps(obj) + "\n")


def write(obj, fn):
    with open(fn, "w") as f:
        json.dump(obj, f, indent=4)


def iread(fn):
    with open(fn) as f:
        for i, line in enumerate(f):
            try:
                yield json.loads(line)
            except json.decoder.JSONDecodeError as e:
                msg = "JSON-L parsing error in line number {} in the jsonl file".format(i)
                raise json.decoder.JSONDecodeError(msg, line, e.pos)


def iwrite(obj, fn):
    with open(fn, "w") as f:
        for chunk in obj:
            f.write(json.dumps(chunk) + "\n")



import inspect
import os
import glob2


def glob(path):
    return glob2.glob(os.path.expanduser(path))


def get_just_env_path():
    return os.environ.get("JUST_PATH")


def find_just_path(base=None, max_depth=5):
    if base is None:
        base = os.path.dirname(os.path.abspath(inspect.stack()[1][1]))
    for depth in range(max_depth):
        prefix = "../" * depth
        just_file = os.path.join(base, prefix, ".just")
        if os.path.isfile(just_file):
            print("just_file", os.path.abspath(os.path.dirname(just_file)))
            return os.path.abspath(os.path.dirname(just_file))
    return None


def get_likely_path():
    # try:
        #main_path = os.path.abspath(sys.modules['__main__'].__file__)
    # except AttributeError:
    main_path = os.path.realpath('__file__')
    return os.path.dirname(main_path)


def get_just_path():
    just_path = get_just_env_path()
    just_path = just_path if just_path is not None else find_just_path()
    just_path = just_path if just_path is not None else find_just_path(".")
    just_path = just_path if just_path is not None else get_likely_path()
    return just_path


def make_path(filename):
    just_path = get_just_path()
    return os.path.join(just_path, os.path.expanduser(filename))



def read(fname):
    with open(fname) as f:
        return f.read()


def iread(fname):
    with open(fname) as f:
        for line in f:
            yield line.rstrip("\n")


def write(obj, fname):
    with open(fname, "w") as f:
        f.write(obj)



import time
import json


def _retry(request_fn, max_retries, delay_base, raw, kwargs):
    from requests import RequestException
    tries = 0
    if 'headers' not in kwargs:
        kwargs["headers"] = {}
    if 'User-Agent' not in kwargs:
        kwargs["headers"]['User-Agent'] = 'Just Agent 1.0'
    while tries < max_retries:
        try:
            r = request_fn(**kwargs)
            break
        except RequestException as e:
            tries += 1
            print("just.requests_", kwargs["url"], "attempt", tries, str(e))
            if tries == max_retries:
                return ""
            time.sleep(delay_base ** tries)
    if raw:
        return r.content
    if "application/json" in r.headers['Content-Type']:
        return r.json()
    return r.text


def get(url, params=None, max_retries=5, delay_base=3, raw=False, **kwargs):
    import requests
    kwargs['url'] = url
    kwargs['params'] = json.dumps(params) if params else ''
    result = _retry(requests.get, max_retries, delay_base, raw, kwargs)
    return result


def post(url, params=None, data=None, max_retries=5, raw=False, delay_base=3, **kwargs):
    import requests
    kwargs['url'] = url
    kwargs['params'] = json.dumps(params) if params else ''
    kwargs['data'] = json.dumps(data) if data else ''
    result = _retry(requests.post, max_retries, delay_base, raw, kwargs)
    return result
